module Main (main) where

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, )
import Foreign.Ptr (Ptr, )

import qualified System.IO as IO

import Data.Word (Word32, )

import Control.Monad (liftM2, void, )


waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t =
   A.sub (valueOf 1) =<<
   A.mul (valueOf 2) t

osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSaw phase) (SoV.incPhase freq phase)


waveRect :: Value Float -> CodeGenFunction r (Value Float)
waveRect t = do
   r <- LLVM.cmp LLVM.CmpGT (valueOf (0.5 :: Float)) t
   LLVM.select r (valueOf 1) (valueOf (-1))


osciRect ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciRect freq phase =
   liftM2 (,) (waveRect phase) (SoV.incPhase freq phase)


moduleWaveForm :: (Value Float -> Value Float -> CodeGenFunction Float (Value Float, Value Float))
  -> CodeGenModule (Function (Float -> Word32 -> Ptr Float -> IO Float))
moduleWaveForm f =
   LLVM.createFunction LLVM.ExternalLinkage $ \ freq size ptr -> do
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- f freq phase0
         LLVM.store y ptri
         return phase1
      LLVM.ret (s :: Value Float)


renderSaw :: IO ()
renderSaw = do
   fill <- simpleFunction (moduleWaveForm osciRect)
   let renderSeq = [(0.03, 100000), (0.02, 100000), (0.01, 100000)]
   IO.withFile "saw.f32" IO.WriteMode $ \h ->
      mapM_ (\ (freq, len) ->
          allocaArray len $ \ ptr ->
             fill freq (fromIntegral len) ptr >>
             IO.hPutBuf h ptr (len*St.sizeOf(undefined::Float))) renderSeq


main :: IO ()
main = do
   LLVM.initializeNativeTarget
   renderSaw
