module Main (main) where

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, )
import Foreign.Ptr (Ptr, )

import qualified System.IO as IO

import Data.Word (Word32, )

import Control.Monad (liftM2, void, )


waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t =
   A.sub (valueOf 1) =<<
   A.mul (valueOf 2) t


waveRect :: Value Float -> CodeGenFunction r (Value Float)
waveRect t = do
   right <- A.cmp LLVM.CmpGT t (valueOf 0.5)
   LLVM.select right (valueOf 0.0) (valueOf 1.0)

waveSine :: Value Float -> CodeGenFunction r (Value Float)
waveSine t = do
   rad <- A.mul (valueOf 2) =<< A.mul t =<< A.pi
   A.sin rad 
osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSine phase) (SoV.incPhase freq phase)



moduleSaw :: CodeGenModule (Function (Word32 -> Float -> Ptr Float -> IO Float))
moduleSaw =
   LLVM.createFunction LLVM.ExternalLinkage $ \ size freq ptr -> do
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- osciSaw freq phase0
         LLVM.store y ptri
         return phase1
      LLVM.ret (s :: Value Float)

writeSawCode :: IO ()
writeSawCode = do
   m <- LLVM.newModule
   void $ LLVM.defineModule m moduleSaw
   LLVM.writeBitcodeToFile "saw.bc" m

renderSaw :: IO ()
renderSaw = do
   fill <- simpleFunction moduleSaw
   let toneX h freq len = 
           allocaArray len $ \ ptr ->
             fill (fromIntegral len) freq ptr  >>
             IO.hPutBuf h ptr (len*St.sizeOf(undefined::Float))
   IO.withFile "saw.f32" IO.WriteMode $ \h -> 
       do
          toneX h 0.01 40000
          toneX h 0.05 40000
          toneX h 0.02 40000



main :: IO ()
main = do
   LLVM.initializeNativeTarget
   writeSawCode
   renderSaw
