module Main (main) where

import qualified LLVM.Extra.ScalarOrVector as SoV
import qualified LLVM.Extra.Arithmetic as A
import LLVM.Extra.Control (arrayLoop, )

import qualified LLVM.Core as LLVM
import LLVM.Core (Value, valueOf, Function, CodeGenModule, CodeGenFunction, )
import LLVM.ExecutionEngine (simpleFunction, )

import qualified Foreign.Storable as St
import Foreign.Marshal.Array (allocaArray, advancePtr )
import Foreign.Ptr (Ptr, )

import qualified System.IO as IO

import Data.Word (Word32, )

import Control.Monad (liftM2, void, forM_)

import qualified System.Process as System


waveSaw :: Value Float -> CodeGenFunction r (Value Float)
waveSaw t =
   A.sub (valueOf 1) =<<
   A.mul (valueOf 2) t

osciSaw ::
   Value Float -> Value Float ->
   CodeGenFunction r (Value Float, Value Float)
osciSaw freq phase =
   liftM2 (,) (waveSaw phase) (SoV.incPhase freq phase)


fromHertz :: Value Float -> CodeGenFunction r (Value Float)
fromHertz t = A.fdiv t (valueOf sampleRate)

moduleSaw :: CodeGenModule (Function (Float -> Word32 -> Ptr Float -> IO Float))
moduleSaw =
   LLVM.createFunction LLVM.ExternalLinkage $ \freq size ptr -> do
      freq2 <- fromHertz freq
      s <- arrayLoop size ptr (valueOf 0) $ \ ptri phase0 -> do
         (y, phase1) <- osciSaw freq2 phase0
         LLVM.store y ptri		-- 
         return phase1
      LLVM.ret (s :: Value Float)	-- Letzte phase

sampleRate :: Float
sampleRate = 44100

writeSawCode :: IO ()
writeSawCode = do
   m <- LLVM.newModule
   void $ LLVM.defineModule m moduleSaw
   LLVM.writeBitcodeToFile "saw.bc" m

renderSaw :: IO ()
renderSaw = do
   fill <- simpleFunction moduleSaw		-- create function from LLVM
   IO.withFile "saw.f32" IO.WriteMode $ \h ->
      let (total, notes) = toBuffers melody

      in  allocaArray total $ \ ptr -> do   	-- Reserviere Speicher und fuelle Buffer
             forM_ notes $ \(freq,(pos,len)) -> do
                 fill freq (fromIntegral len) (advancePtr ptr $ fromIntegral pos)
             IO.hPutBuf h ptr (total*St.sizeOf(undefined::Float))


toBuffers :: [(Float, Float)] -> (Int, [(Float, (Word32,Word32))])
toBuffers melody = (fromIntegral $ last pos, zip freqs (zip pos dts))
    where
    pos   = scanl (+) 0 dts
    dts   = [floor (dt * sampleRate) | (_,dt) <- melody]
    freqs = [220*2**(x/12) | (x,_) <- melody]


melody :: [(Float, Float)]
melody = [(0,0.2),(7,0.4),(12,0.4),(11,0.2),(12,0.6),(15,0.4),(12,0.4)]


main :: IO ()
main = do
   LLVM.initializeNativeTarget
   writeSawCode
   renderSaw
   System.system $ "play -r " ++ show sampleRate ++ " saw.f32"
   return ()


